/********************************************************************************************
Copyright (c) 2014-2016, Sven Reimer,
Copyright (c) 2017-2020, Tobias Paxian

dPermission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
********************************************************************************************/

#include <assert.h>
#include <cmath> 
#include <iomanip>


#include "../../solver-proxy/SATSolverProxy.h"
#include "../Softclause.h"
#include "cascade.h"
#include "bucket.h"
#include "totalizerencodetree.h"
#include "dgpw.h"
#include "multiplecascade.h"
#include "sorter.h"
#include "timemeasurement.h"
#include "timevariables.h"

#include <algorithm>
#include <chrono>
#include <iostream>
#include <iterator>
#include <random>
#include <vector>

namespace Pacose {
namespace DGPW {
DGPW::DGPW(Pacose *pacose, VeriPbProofLogger* vPL)
    : _sumOfSoftWeights(),
      _timeVariables(nullptr),
      _softClauses(),
      _sorterTree(),
      _solver(pacose->_satSolver),
      _dgpwSetting(&pacose->_settings),
      _pacose(pacose),
      _softClausesFixed(false),
      _maxSorterDepth(0),
      _maxsatResult(UNKNOW),
      _mainCascade(nullptr),
      _mainMultipleCascade(nullptr),
      _greatestCommonDivisor(1),
      _currentBucketForTare(0),
      _satWeight(0),
      _moreThanTwoWeights(false),
      //    _topWeight(pacose->_top),
      _minWeight(1),
      _maxWeight(0),
      _hasEncoding(false),
      _resultUnknown(false),
      _clausesBefore(pacose->_nbClauses),
      _variablesBefore(pacose->_nbVars),
      _binaryClausesBefore(0),
      _ternaryClausesBefore(0),
      _addedClauses(0),
      _addedVariables(0),
      _addedBinaryClauses(0),
      _addedTernaryClauses(0),
      _binaryTopClauses(0),
      _ternaryTopClauses(0),
      _binaryBottomClauses(0),
      _ternaryBottomClauses(0),
      _solverCalls(0),
      _approxSolverCalls(0),
      _highestBucketMultiplicator(0),
      _currentBucketMultiplicator(0),
      _estimatedSatWeight(0),
      _diffToSatWeight(0),
      //    _collectedAssumptions(),
      _externalAssumptions({}),
      _fixedAssumptions({}),
      _lastResult(0),
      vPL(vPL) {
  _sorterTree.resize(1);
  }

void DGPW::IncrementalReset() {
  *_optimum = -1;
  _maxsatResult = UNKNOW;
  _currentBucketForTare = 0;
  _satWeight = 0;
  _moreThanTwoWeights = true;
  _minWeight = 1;
  _maxWeight = 0;
  _resultUnknown = false;
  _clausesBefore = _pacose->_nbClauses;
  _variablesBefore = _pacose->_nbVars;
  _solverCalls = 0;
  _approxSolverCalls = 0;
  _currentBucketMultiplicator = 1;
  _estimatedSatWeight = 0, _diffToSatWeight = 0;
  _lastResult = 0;
  _mainCascade->IncrementalReset();
}

DGPW::~DGPW() {
  for (uint32_t i = 0; i != _sorterTree.size(); ++i) {
    for (uint32_t j = 0; j != _sorterTree[i].size(); ++j) {
      delete _sorterTree[i][j];
    }
  }
  delete _timeVariables;
  delete _mainCascade;
  delete _mainMultipleCascade;

  // the following are destroyed elsewhere!
  //  SATSolverProxy *_solver;
  //  Settings *_dgpwSetting;
  //  Pacose *_pacose;
}

/********** Interface to Glucose and reimplemented functions from AntomBase
 * **********/

// Creates and returns a fresh variable index not used so far.
uint32_t DGPW::NewVariable(void) {
  _addedVariables++;
  return static_cast<uint32_t>(_solver->NewVariable());
  //    Glucose::Var newVar = _glucose->newVar(true, true);
  //    return static_cast<uint32_t>(newVar) + 1;
}

// TODO: optimize -> change interface to Glucose data structures
bool DGPW::AddClause(std::vector<uint32_t> &clause, uint32_t lbd) {
  //  std::cout << "Clause: ";
  //  for (auto literal : clause) {
  //    std::cout << literal << " ";
  //  }
  //  std::cout << std::endl;
  _addedClauses++;
  return _solver->AddClause(clause);
  //    Glucose::vec<Glucose::Lit> newClause;
  //    for (uint32_t lit : clause )
  //      {
  //        assert(lit > 1);
  //        Glucose::Lit newLit = Glucose::toLit(static_cast<int>(lit - 2));
  //        newClause.push(newLit);
  //      }
  //    return _glucose->addClause_(newClause);
}

//  bool DGPW::AddClause(Glucose::vec<Glucose::Lit>& clause)
//  {
//    return _glucose->addClause_(clause);
//  }

uint32_t DGPW::GetLastResult() { return _lastResult; }

bool DGPW::AddUnit(uint32_t lit) {
  std::vector<uint32_t> unitclause;
  unitclause.push_back(lit);
  //  std::cout << " "
  //               "                          UNIT CLAUSE: "
  //            << lit << std::endl;

  return AddClause(unitclause);
}

uint32_t DGPW::Solve(void) {
  std::cout << "c SOLVE(void)!" << std::endl;
  _solver->ClearAssumption();

  if (!_fixedAssumptions.empty()) {
    std::cout << "_fixedAssumptions not empty: " << _fixedAssumptions.size() << std::endl;
    _solver->AddAssumptions(_fixedAssumptions);
  }
  _lastResult = _solver->Solve();
  if(_lastResult == SAT){
    _pacose->CalculateLocalSATWeight();
    _pacose->CertWriteModel(_mainCascade->_structure.back()->_sorter->_outputTree->_tares, lbTderivedfor);
  }
  _solverCalls++;
  return _lastResult;
}

uint32_t DGPW::Solve(std::vector<uint32_t> &assumptions) {
   std::cout << "c SOLVE(assumptions)!"
             << std::endl;
  _solver->ClearAssumption();
  if (!_fixedAssumptions.empty()) {
    _solver->AddAssumptions(_fixedAssumptions);
  }
  _solver->AddAssumptions(assumptions);
  _lastResult = _solver->Solve();
  if(_lastResult == SAT){
    _pacose->CalculateLocalSATWeight();
    _pacose->CertWriteModel(_mainCascade->_structure.back()->_sorter->_outputTree->_tares, lbTderivedfor);
    // _satWeight = _pacose->_localSatWeight;
    // *_optimum  = _pacose->_localUnSatWeight;
    // _mainCascade->_satWeight = _satWeight;
  }
  _solverCalls++;
  return _lastResult;
  //    std::cout << "result: " << rv << std::endl;
  //    return rv;
}

// TODO: optimize!
// copies whole model vector
std::vector<uint32_t> DGPW::Model(void) const {
  //    std::cout << __PRETTY_FUNCTION__ << "  x  " << std::endl;

  //    _solver->ClearAssumptions();
  //    std::cout << _solver->Solve() << std::endl;

  std::vector<uint32_t> model;

  for (int32_t i = 0; i < static_cast<int32_t>(_solver->GetNumberOfVariables());
       ++i) {
    model.push_back(_solver->GetModel(i));
  }
  return model;
}

// TODO: Check this for correctness!
// TODO: Try to use this interface whenever possible
uint32_t DGPW::Model(uint32_t var) const {
  //      std::cout << __PRETTY_FUNCTION__ << "  y  " << std::endl;
  //    assert(var>0);
  //      std::cout << "_solver->GetLastResult(): " << _solver->GetLastResult()
  //      << std::endl;
  if (_lastResult != 10) return 0;

  assert(var <= Variables());
  uint32_t value = _solver->GetModel(static_cast<int>(var));
  assert(value <= 2 * Variables() + 1);
  assert(value <= 2 * Variables() + 1);
  return value;
}

uint32_t DGPW::Variables() const { return _solver->GetNumberOfVariables(); }

uint32_t DGPW::Clauses() const { return _solver->GetNumberOfClauses(); }

uint32_t DGPW::StaticClauses() const {
  //    return static_cast<uint32_t>(_glucose->nClauses());
  return Clauses();
}

uint32_t DGPW::CurrentBinaryClauses() const {
  // TODO: implement me!
  return 0;
}
uint32_t DGPW::CurrentTernaryClauses() const {
  // TODO: implement me!
  return 0;
}

Settings *DGPW::GetSetting() { return _dgpwSetting; }

bool DGPW::HasEncoding() { return _hasEncoding; }

/********** End: Interface to Glucose and reimplemented functions from AntomBase
 * **********/

bool DGPW::AddWeightedSoftClause(std::vector<uint32_t> &clause,
                                 uint64_t weight) {
  SoftClause *sclaus = new SoftClause(0, clause, weight);
  _softClauses.push_back(sclaus);

  return true;
}

void DGPW::SetSoftClauseVector(std::vector<SoftClause *> *softClauses) {
  _softClauses = *softClauses;
}

// If sorter == nullptr, we want to count for all soft clauses
uint64_t DGPW::CountSatisfiedSoftClauses(Sorter *sorter,
                                         const std::vector<uint32_t> &model) {
  //    std::cout << __PRETTY_FUNCTION__ << std::endl;

  std::vector<SoftClause *> softclauses;
  if (sorter == nullptr) {
    softclauses = _softClauses;
  } else {
    softclauses = sorter->GetSoftClauses();
  }

  uint64_t result = CountSatisfiedSoftClauses(softclauses, model);

  if (sorter != nullptr) {
    sorter->SetMinSatisfied(result);
  }
  return result;
}

uint64_t DGPW::CountSatisfiedSoftClauses(std::vector<SoftClause *> softclauses,
                                         const std::vector<uint32_t> &model) {
  if (_dgpwSetting->verbosity > 4) {
    std::cout << __PRETTY_FUNCTION__ << std::endl;
    std::cout << "SC.size: " << softclauses.size() << std::endl;
    //        std::cout << "Clauses.size: " << Clauses() << std::endl;
  }
  uint64_t result(0);
  // Proceed all soft clauses
  for (uint32_t i = 0; i != softclauses.size(); ++i) {
    uint32_t relaxlit = softclauses[i]->relaxationLit;

    //    if( model[relaxlit>>1] == relaxlit )
    if (Model(relaxlit >> 1) == relaxlit) {
      std::vector<uint32_t> clause(softclauses[i]->clause);
      uint32_t pos = 0;
      for (; pos != clause.size(); ++pos) {
        // clause satisfied without trigger?
        if (Model(clause[pos] >> 1) == clause[pos]) {
          result += softclauses[i]->weight;
          softclauses[i]->lastassignment = relaxlit ^ 1;
          //            std::cout << "++" << result;
          break;
        }
      }
      if (pos == clause.size()) {
        softclauses[i]->lastassignment = relaxlit;
      }
    } else if (Model(relaxlit >> 1) != 0) {
      assert(Model(relaxlit >> 1) == (relaxlit ^ 1));
      softclauses[i]->lastassignment = relaxlit ^ 1;
      result += softclauses[i]->weight;
      //        std::cout << " ++relax sat " << result;
    }
  }
  return result;
}

// Weigthed maxsat
// ____________________________________________________________________________________________________________________
uint32_t DGPW::MaxSolveIncremental() {
  IncrementalReset();

  // _satWeight = 0;
  // _minWeight = 1;
  // _maxWeight = 0;

  uint32_t currentresult = Solve();

  if (currentresult != SAT) return currentresult;

  CalculateOverallOptimum(_satWeight, true);

  currentresult = _mainCascade->Solve(true, true);

  return currentresult;
}

// Weigthed maxsat
// ____________________________________________________________________________________________________________________
uint32_t DGPW::MaxSolveWeightedPartial(int64_t &optimum) {
  std::vector<uint32_t> externalAssumptions = {};
  uint32_t result = MaxSolveWeightedPartial(externalAssumptions, optimum);

  return result;
}

// ____________________________________________________________________________________________________________________
uint32_t DGPW::MaxSolveWeightedPartial(
    std::vector<uint32_t> &externalAssumptions, int64_t &optimum) {
  //    std::cout << __PRETTY_FUNCTION__ << std::endl;
  //    _dgpwSetting->Print();

  // to change optimum value with another function called from cascade and
  // bucket
  _optimum = &optimum;
  _hasEncoding = false;
  //    _dgpwSetting->application = WEIGHTEDMAXSAT;

  _timeVariables = new TimeVariables();
  TimeMeasurement totalTime(&_timeVariables->total, true);

  uint32_t currentresult(SAT);
  if (_dgpwSetting->checkIfSolutionIsUnique &&
      _dgpwSetting->currentCascade.iteration > 0) {
    if (_dgpwSetting->currentCascade._onlyWithAssumptions) {
      std::cout << "c BAD PARAMETER COMBINATION!!!" << std::endl;
      exit(1);
    }
    _dgpwSetting->currentCascade._onlyWithAssumptions = true;
  }

  // uint32_t averageBucketEntries(0);

  //#ifndef NDEBUG
  uint64_t softWeights = 0;
  std::for_each(_softClauses.begin(), _softClauses.end(),
                [&](SoftClause *SC) { softWeights += SC->weight; });
  _sumOfSoftWeights = softWeights;

  //	std::cout << "soft weights: " << softWeights << " sumofweights: " <<
  //_sumOfSoftWeights << std::endl;
  //    assert(softWeights == _sumOfSoftWeights);
  //#endif
  if (_sumOfSoftWeights == 1) {
  }

  //  std::cout << "c clauses before.........: " << _clausesBefore << std::endl;
  if (_dgpwSetting->verbosity > 0) {
    std::cout << "c #softclauses...........: " << _softClauses.size()
              << std::endl;
    std::cout << "c #sum of SoftWeights....: "
              << _sumOfSoftWeights * _greatestCommonDivisor << std::endl;
  }
  //    if (!_dgpwSetting->solveAtFirst)
  //      std::cout << "c #sum of SoftWeights....: " << _sumOfSoftWeights *
  //      _greatestCommonDivisor << std::endl;

  if (_dgpwSetting->onlyByTares ||
      _dgpwSetting->mcDivideStrategy != SOLVEINNORMALCASCADEMODE) {
    _dgpwSetting->encodeStrategy = ENCODEONLYIFNEEDED;
  }

  if (_satWeight == _sumOfSoftWeights) {
    _pacose->wbSortAndFilter();

    std::cout << "c local o 0" << std::endl;
    std::cout << "c All SoftClauses of the current gbmo level are satisfiable!" << std::endl;
    return SAT;
  }

  // calc and set all trivially conflicting softclauses
  // TOASK: maybe better seperately for each bucket - to know boundaries of
  // bucket! together with mode of solving bucket parts to get max
  // CheckAllWeightedConflictingSoftclauses();
  if (_dgpwSetting->mcDivideStrategy != SOLVEINNORMALCASCADEMODE) {
    _dgpwSetting->encodeStrategy = ENCODEONLYIFNEEDED; // cone of influence encoding
    _mainMultipleCascade = new MultipleCascade(
        this, _dgpwSetting->onlyByTares, _dgpwSetting->tareCascadeOnlyByTares,
        _dgpwSetting->cascadeDivider, _dgpwSetting->maxBucketSize,
        _dgpwSetting->nOfCasc, _dgpwSetting->interimResult, _sumOfSoftWeights);
    //        std::cout << "Cascade Divider: " << _cascadeDivider << std::endl;
    if (_mainMultipleCascade->DivideAndConnectCascades(
            _dgpwSetting->mcDivideStrategy, &_softClauses)) {
      currentresult = _mainMultipleCascade->Solve();

      if (currentresult != UNKNOW) currentresult = SAT;
    }
  } else {
    _dgpwSetting->cascadeDivider = 0;
  }

  // no else if, because if _cascadeDivider returns only one cascade then
  // _cascadeDivider == 0
  if (_dgpwSetting->cascadeDivider == 0) {
    _clausesBefore = StaticClauses();
    _mainMultipleCascade = nullptr;
    //        std::cout << "_dgpwSetting->onlyByTares: " <<
    //        _dgpwSetting->onlyByTares << std::endl;
    _mainCascade = new Cascade(this, nullptr, _dgpwSetting->onlyByTares);

    //        std::cout << "PartitionStrategy: " <<
    //        _dgpwSetting->partitionStrategy << std::endl;

    // Only Tare Variables are generated in here!
    _mainCascade->Fill(&_softClauses, _dgpwSetting->partitionStrategy,
                       _dgpwSetting->encodeStrategy);
    _hasEncoding = true;

    // if (_antomSetting->verbosity > 2 && _antomSetting->base > 2)
    //{
    // std::cout << "How often is Softclause in Bucket, because of base " <<
    // _baseMode << std::endl; DumpBucketStructure();
    //}

    if (_dgpwSetting->cascadeDivider == 0) {
      if (_dgpwSetting->encodeStrategy == ENCODEALL) {
        _clausesBefore = StaticClauses();
      }
      if (!_mainCascade->Encode()) {
        return UNKNOW;
      }
      //      if (_dgpwSetting->encodeStrategy == ENCODEALL) {
      //        std::cout << "c #clauses of coding.....: "
      //                  << StaticClauses() - _clausesBefore << std::endl;
      //        std::cout << "c #variables of coding...: "
      //                  << Variables() - _variablesBefore << std::endl;
      //      }
    }

    // TOBI: SetDecisionStrategiesForCascadeModel!!!!
    // SetDecisionStrategiesForMaxSAT();
    currentresult =
        _mainCascade->Solve(_dgpwSetting->currentCascade._onlyWithAssumptions,
                            _dgpwSetting->currentCascade._solveTares);
  }

  //  if (_dgpwSetting->encodeStrategy != ENCODEALL &&
  //  _dgpwSetting->onlyByTares) {
  //    std::cout << "c #clauses of coding.....: " << _addedClauses <<
  //    std::endl; std::cout << "c #variables of coding...: " << _addedVariables
  //    << std::endl;
  //  }
  if (currentresult == UNKNOW) {
    return UNKNOW;
  }

  // if there is more than one cascade, we have to calculate
  // the OverallOptimum again to sum up the weights.
  //    CalculateOverallOptimum(_satWeight, true);
  optimum = optimum * _greatestCommonDivisor;

  if (_dgpwSetting->checkIfSolutionIsUnique &&
      _dgpwSetting->currentCascade.iteration > 0) {
    if (optimum == 0) {
      // special case 1 - all SCs are satisfied!
      for (auto SC : _softClauses) {
        AddUnit(SC->relaxationLit ^ 1);
      }
      //            currentresult = Solve();
      //            if (currentresult != SAT) {
      //                std::cout << "c Strange result - shouldn't be possible!"
      //                << std::endl; exit(1);
      //            }
    } else if (_satWeight == 0) {
      // special case 2 - no SC could be satisfied
      for (auto SC : _softClauses) {
        AddUnit(SC->relaxationLit);
      }
      //            currentresult = Solve();
      //            if (currentresult != SAT) {
      //                std::cout << "c Strange result - shouldn't be possible!"
      //                << std::endl; exit(1);
      //            }
    } else {
      std::vector<uint32_t> clause;
      for (auto SC : _softClauses) {
        clause.push_back(Model(SC->relaxationLit >> 1) ^ 1);
      }
      uint32_t relaxLit = NewVariable() << 1;
      clause.push_back(relaxLit);
      AddClause(clause);

      std::vector<uint32_t> assumptions = GetLastAssumptions();

      if (assumptions.size() == 0) {
        std::cout << "c NO ASSUMPTIONS - ERROR!!!" << std::endl;
        exit(1);
      }
      assumptions.push_back(relaxLit ^ 1);

      uint32_t currentresult = Solve(assumptions);

      if (currentresult != SAT) {
        std::cout << "c SOLUTION IS UNIQUE! Add this solution as hard clauses "
                     "to the problem!"
                  << std::endl;
        AddUnit(relaxLit);
        clause.pop_back();
        // add solution as unit clauses
        for (auto lit : clause) {
          AddUnit(lit ^ 1);
        }
        // we don't need the encoding anymore - throw it away!!
        for (auto ass : assumptions) {
          AddUnit(ass ^ 1);
        }
        currentresult = Solve();
        if (currentresult != SAT) {
          std::cout << "c Strange result - shouldn't be possible!" << std::endl;
          exit(1);
        }
      } else {
        std::cout << "c THERE ARE MULTIPLE SOLUTIONS!" << std::endl;
        assumptions.pop_back();
        AddUnit(relaxLit);
        for (auto ass : assumptions) {
          std::cout << ass << ", ";
          AddUnit(ass);
        }
      }
    }
  }

  if (_dgpwSetting->verbosity > 0) {
    if (!_dgpwSetting->formulaIsDivided) {
      std::cout << "c s SATISFIABLE" << std::endl;
      //    std::cout << "o " << optimum << std::endl;
    } else {
      if (currentresult == SAT) {
        assert(std::cout << "c assertion Solver Call in MaxSolveWeightedPartial: " << std::endl && Solve() == SAT);
        std::cout << "c currently SAT" << std::endl;
        //      std::cout << "c local o " << optimum << std::endl;
      } else if (currentresult == 20) {
        std::cout << "c currently UNSAT" << std::endl;
        //      std::cout << "c local o " << optimum << std::endl;
      }
    }
  }

  if (_satWeight == _sumOfSoftWeights &&
      !_dgpwSetting->currentCascade._onlyWithAssumptions) {
    // add relaxation literals as UnitClauses!
    // for (auto sc : _softClauses) {
    //   // If all soft clauses can be satisfied, we satisfy them by adding unit clause that propagates them.
    //   vPL->write_comment("All soft clauses can be satisfied and are therefore fixed by unit clauses.");
    //   vPL->unchecked_assumption_unit_clause(sc->relaxationLit ^ 1);
    //   AddUnit(sc->relaxationLit ^ 1);
    // }
    std::cout << "c All soft clauses can be satisfied and are therefore fixed by unit clauses." << std::endl;
    FixAllSoftClauses();
    // for (uint32_t i = 0; i < (*_pacose->_actualSoftClauses).size(); i++) {
    //   vPL->write_comment("All soft clauses can be satisfied and are therefore fixed by unit clauses.");
    //   vPL->unchecked_assumption_unit_clause((*_pacose->_actualSoftClauses)[i]->relaxationLit ^ 1);
    //   AddUnit((*_pacose->_actualSoftClauses)[i]->relaxationLit ^ 1);
    //   // TODO DIETER -- now the SCs are removed from the vector as well!
    //   (*_pacose->_actualSoftClauses).erase((*_pacose->_actualSoftClauses).begin() + i);
    // }
    currentresult = Solve();

    if (_dgpwSetting->verbosity > 0)
      std::cout << "c All SoftClauses are Satisfiable!" << std::endl;
    assert(currentresult == SAT);
  }
  if (_dgpwSetting->verbosity > 0)
    std::cout << "c number of variables: " << Variables() << std::endl;
  return currentresult;
}

void DGPW::FixAllSoftClauses() {
  std::cout << __PRETTY_FUNCTION__ << std::endl;
  if (_softClausesFixed)
    return;
  assert(std::cout << "c assertion Solver Call in MaxSolveWeightedPartial2: " << std::endl && Solve() == SAT);
  vPL->write_comment("FixAllSoftClauses: All soft clauses can be satisfied and are therefore fixed by unit clauses.");
  std::vector<uint32_t> varstoweaken;

  for (uint32_t i = 0; i < (*_pacose->_actualSoftClauses).size(); i++) {
    
    vPL->rup_unit_clause((*_pacose->_actualSoftClauses)[i]->relaxationLit ^ 1);
    vPL->move_to_coreset(-1, true);
    vPL->copy_constraint(-1);
    AddUnit((*_pacose->_actualSoftClauses)[i]->relaxationLit ^ 1);

    bool litInObj =  vPL->remove_objective_literal((*_pacose->_actualSoftClauses)[i]->relaxationLit);
    if(litInObj){ // Literal might already be removed by somewhere else.
      std::vector<uint32_t> litsObjU = {(*_pacose->_actualSoftClauses)[i]->relaxationLit};
      std::vector<signedWght> wghtsObjU = {-static_cast<signedWght>((*_pacose->_actualSoftClauses)[i]->originalWeight)} ;
      vPL->write_objective_update_diff(litsObjU, wghtsObjU);
      varstoweaken.push_back(variable((*_pacose->_actualSoftClauses)[i]->relaxationLit));
    }  
  }

  if(varstoweaken.size() > 0){
    vPL->start_intCP_derivation(vPL->get_model_improving_constraint());
    for(uint32_t var : varstoweaken){
      vPL->intCP_weaken(var);
    }
    constraintid newmic = vPL->end_intCP_derivation();

    // cuttingplanes_derivation cpder = vPL->CP_constraintid(vPL->get_model_improving_constraint());
    // for(uint32_t var : varstoweaken){
    //   cpder = vPL->CP_weakening(cpder, var);
    // }
    // constraintid newmic = vPL->write_CP_derivation(cpder);
    vPL->update_model_improving_constraint(newmic);
    vPL->check_model_improving_constraint(newmic);
  }
  // TODO DIETER -- now the SCs are removed from the vector as well!
  (*_pacose->_actualSoftClauses).clear();
  // std::cout << "(*_pacose->_actualSoftClauses).size(): " << (*_pacose->_actualSoftClauses).size() << std::endl;
  _softClausesFixed = true;
  assert(std::cout << "c assertion Solver Call in FixAllSoftClauses: " << std::endl && Solve() == SAT);
}

void DGPW::SetInitialAssumptions(std::vector<uint32_t> assumptions) {
  _externalAssumptions = assumptions;
}

StructureInfo DGPW::AnalyzeStructure() {
  // assert(_sumOfSoftWeights > _topWeight);
  bool maxIsMin = (_maxWeight == _minWeight);
  bool withHC = Clauses() != 0;
  bool minIsSet = _minWeight != ((uint64_t)-1);
  bool maxIsSet = _maxWeight != 0;
  // is SAT formula
  bool onlyHC = !minIsSet && !maxIsSet && withHC;
  // is MaxSAT formula -- NOT PARTIAL
  bool onlySCwithOneWeight = !_moreThanTwoWeights && maxIsMin && !withHC;
  // is partial MaxSAT formula - Set _minWeight to one
  bool oneHCWeightAndSCWeight = maxIsMin && withHC;
  // if sum (_minWeight) < _maxWeight --> is partial MaxSAT formula
  bool onlySCwithTwoWeights =
      !_moreThanTwoWeights && !maxIsMin && minIsSet && maxIsSet && !withHC;
  // Calculate the greatest common divisor of all softclauseweights.
  uint64_t greatestCommonDivisor = _minWeight;

  for (uint32_t ind = 0; ind < _softClauses.size(); ++ind) {
    if (greatestCommonDivisor == 1) break;
    greatestCommonDivisor =
        GreatestCommonDivisor(greatestCommonDivisor, _softClauses[ind]->weight);
  }
  _greatestCommonDivisor = greatestCommonDivisor;

  if (_dgpwSetting->verbosity > 2) {
    std::cout << std::setw(30) << "_topWeight " << _topWeight << std::endl;
    std::cout << std::setw(30) << "_minWeight " << _minWeight << std::endl;
    std::cout << std::setw(30) << "_maxWeight " << _maxWeight << std::endl;
    std::cout << std::setw(30) << "_moreThanTwoWeights " << _moreThanTwoWeights
              << std::endl;
    std::cout << std::setw(30) << "maxIsMin " << maxIsMin << std::endl;
    std::cout << std::setw(30) << "withHC " << withHC << std::endl;
    std::cout << std::setw(30) << "minIsSet " << minIsSet << std::endl;
    std::cout << std::setw(30) << "maxIsSet " << maxIsSet << std::endl;
    std::cout << std::setw(30) << "onlyHC " << onlyHC << std::endl;
    std::cout << std::setw(30) << "onlySCwithOneWeight " << onlySCwithOneWeight
              << std::endl;
    std::cout << std::setw(30) << "oneHCWeightAndSCWeight "
              << oneHCWeightAndSCWeight << std::endl;
    std::cout << std::setw(30) << "onlySCwithTwoWeights "
              << onlySCwithTwoWeights << std::endl;
    std::cout << std::setw(30) << "_sumOfSoftWeights " << _sumOfSoftWeights
              << std::endl;
  }

  if (onlyHC) {
    return ISSAT;
  }
  if (onlySCwithTwoWeights) {
    uint32_t sumOfMinWeights(0);

    // the sum of all _minWeights has to be smaller than _maxWeight
    // then it can be solved like MaxSAT formula.
    for (uint32_t ind = 0; ind < _softClauses.size(); ++ind) {
      if (_softClauses[ind]->weight != _minWeight) continue;

      sumOfMinWeights += _softClauses[ind]->weight;
    }

    if (sumOfMinWeights < _maxWeight) {
      _sumOfSoftWeights = sumOfMinWeights;
      return CONVERTTOMAXSAT;
    }
  }
  if (oneHCWeightAndSCWeight || onlySCwithOneWeight || oneHCWeightAndSCWeight) {
    return ISMAXSAT;
  }
  //    if (greatestCommonDivisor > 1)
  //      {
  //        _greatestCommonDivisor = greatestCommonDivisor;
  //        std::cout << "c greatest common divisor: " <<
  //        greatestCommonDivisor
  //        << std::endl; return dgpw::DIVWEIGHTSBYDIVISOR;
  //      }
  return ISWEIGHTEDMAXSAT;
}

uint64_t DGPW::GreatestCommonDivisor(uint64_t a, uint64_t b) {
  uint64_t temp;
  while (b > 0) {
    temp = b;
    b = a % b;
    a = temp;
  }
  return a;
}

void DGPW::ConvertFormulaToMaxSAT(uint64_t maxWeight) {
  // configure all _maxWeights as hardclauses
  std::vector<SoftClause *> newSoftClauseVector;
  for (uint32_t ind = 0; ind < _softClauses.size(); ++ind) {
    if (_softClauses[ind]->weight != maxWeight) {
      newSoftClauseVector.push_back(_softClauses[ind]);
      continue;
    }

    AddClause(_softClauses[ind]->clause);
  }

  _softClauses = newSoftClauseVector;
}

void DGPW::DivideAllSoftClausesByFactor(uint64_t factor) {
  for (uint32_t ind = 0; ind < _softClauses.size(); ++ind) {
    _softClauses[ind]->weight /= factor;
  }
  _sumOfSoftWeights /= factor;
}

uint64_t DGPW::CalcGlobalOpt() { return _pacose->CalculateSATWeight(); }

uint64_t DGPW::CalculateOverallOptimum(uint64_t satWeight, bool countAgain) {
  if (_dgpwSetting->verbosity > 3)
    std::cout << __PRETTY_FUNCTION__ << std::endl;

  // struct rusage resources;
  // Cascade and Bucket counts the satisfied SC only for their SC's
  // for an overall result count again.
  if (countAgain) {
    // maybe _maxSatWeight - because it could be a local optima of this Solver
    // call
    satWeight = CountSatisfiedSoftClauses(nullptr, {});
  }

  // std::cout << std::setw(80) <<std::right<< "OPTIMUM: " << *_optimum << std::endl;

  if (*_optimum > static_cast<int64_t>(_sumOfSoftWeights - satWeight) ||
      *_optimum == -1) {
    /*
      if (_antomSetting->verbosity > 0)
      {
      getrusage(RUSAGE_SELF, &resources);
      double timeC = (double) resources.ru_utime.tv_sec + 1.e-6 * (double)
      resources.ru_utime.tv_usec; std::cout << "c " << (timeC -
      _control->GetStartTime()) << "s" << std::endl;
      }
    */
    // better actualize satWeight once at the end.
    _satWeight = satWeight;
    //    std::cout << "_sumOfSoftWeights: " << _sumOfSoftWeights
    //              << "  satWeight: " << satWeight << std::endl;
    *_optimum = static_cast<int64_t>(_sumOfSoftWeights - satWeight);


    // std::cout <<std::setw(80) << std::right <<  "FORMULA IS DIVIDED? " << _dgpwSetting->formulaIsDivided << std::endl;
    if (!_dgpwSetting->formulaIsDivided) {
      std::cout << "o " << *_optimum * _greatestCommonDivisor << std::endl;
    } else {
      _pacose->CalculateSATWeight();
    }

    if (_dgpwSetting->verbosity > 0)
      std::cout << std::setw(50) << "Calculated Global SATWeight: " << satWeight
                << std::endl;

  } else if (_dgpwSetting->verbosity > 2) {
    /*
      if (_antomSetting->verbosity > 3)
      {
      getrusage(RUSAGE_SELF, &resources);
      double timeC = (double) resources.ru_utime.tv_sec + 1.e-6 * (double)
      resources.ru_utime.tv_usec;
      //std::cout << "c " << (timeC - _control->GetStartTime()) << "s" <<
      std::endl;
      }
    */
    if (!_dgpwSetting->formulaIsDivided)
      std::cout << "o " << *_optimum * _greatestCommonDivisor << std::endl;
    std::cout << std::setw(50) << "Calculated Global SATWeight: " << satWeight
              << std::endl;
  }

  return _satWeight;
}

void DGPW::SetMoreThanTwoWeights(bool val) { _moreThanTwoWeights = val; }

bool DGPW::GetHasMoreThanTwoWeights() { return _moreThanTwoWeights; }

void DGPW::SetTopWeight(uint64_t val) {
  assert(val > 0);
  _topWeight = val;
}

void DGPW::SetMinWeight(uint64_t val) {
  // assert( val < (uint64_t) - 1 );
  _minWeight = val;
}

uint64_t DGPW::GetMinWeight() { return _minWeight; }

void DGPW::SetMaxWeight(uint64_t val) {
  // assert( val > 0 );
  _maxWeight = val;
}

uint64_t DGPW::GetMaxWeight() { return _maxWeight; }

uint64_t DGPW::GetSATWeight() { return _satWeight; }

std::vector<std::pair<uint64_t, uint32_t>> DGPW::GetTareVector(
    uint64_t weightDiff) {
  if (_mainCascade == nullptr) return {};

  return _mainCascade->GetTareVector(weightDiff);
}

std::vector<std::pair<uint64_t, uint32_t>> DGPW::GetWatchdogs(
    uint64_t weightDiff) {
  if (_mainCascade == nullptr) {
    return {};
  }

  return _mainCascade->GetWatchdogs(weightDiff);
}

std::vector<uint32_t> DGPW::GetLastAssumptions() {
  if (_mainCascade == nullptr) {
    return {};
  }
  return _mainCascade->GetLastAssumptions();
}

void DGPW::SetFixedAssumptions(std::vector<uint32_t> fixedAssumptions) {
  _fixedAssumptions = fixedAssumptions;
  //  if (_dgpwSetting->verbosity > 0) std::cout << "c fixed added assumptions:
  //  "; for (auto fa : fixedAssumptions) {
  //    std::cout << fa << " ";
  //  }
  //  std::cout << std::endl;
}

void DGPW::RemoveFixedAssumptions() { _fixedAssumptions.clear(); }

void DGPW::SetHasHardClauses(bool val) { _hasHardClauses = val; }

void DGPW::SetSatWeight(uint64_t val) {
  _satWeight = val;
  //  _minWeight = val;
}

void DGPW::SetUnSatWeight(uint64_t val) {
  *_optimum = static_cast<int64_t>(val);
  //  _minWeight = val;
}

bool DGPW::GetHasHardClauses() { return _hasHardClauses; }

void DGPW::SetGreatestCommonDivisor(uint64_t val) {
  _greatestCommonDivisor = static_cast<int64_t>(val);
}


// --------------- Proof Logging ----------------
void DGPW::GetAllLeavesAndWeights(std::vector<uint32_t>& leaves, std::vector<uint64_t>& weights){
  GetOutputTree()->GetAllLeavesAndWeights(leaves, weights);
}

TotalizerEncodeTree* DGPW::GetOutputTree(){
  return _mainCascade->_structure.back()->_sorter->_outputTree;
}

uint32_t DGPW::GetKopt(){
  return kopt;
}

// constraintid DGPW::GetCxnCCsat(){
//   return _mainCascade->_structure.back()->cxnCCsat;
// }
constraintid DGPW::GetCxnCCunsat(){
  return cxnCCunsat;
}

uint32_t DGPW::GetLitCCunsat(){
  return litCCunsat;
}

uint32_t DGPW::GetP() {
  return _mainCascade->_structure.size() - 1;
}

constraintid DGPW::GetCxnLbT(){
  return cxnlbT;
}
constraintid DGPW::GetCxnUbT(){
  return cxnubT;
}

void DGPW::CombineLeavesForBottomBucket(TotalizerEncodeTree* current,
                                        TotalizerEncodeTree *bb,
                                        TotalizerEncodeTree *tb) {
  if (_dgpwSetting->verbosity > 10)
    std::cout << __PRETTY_FUNCTION__ << std::endl;
  // Note that bb's leaves are already sorted by the recursive call!
  std::sort(tb->_leaves.begin(), tb->_leaves.end());

  uint32_t i = 0, j = 0;

  current->_leaves.reserve(bb->_leaves.size() + tb->_leaves.size());
  if (_dgpwSetting->verbosity > 10)
    std::cout << "Before while loop..." << std::endl;

  while (i < tb->_leaves.size() &&
          j < bb->_leaves
                  .size()) { // Literal in both in bottom bucket and top bucket
    if (tb->_leaves[i] == bb->_leaves[j]) {
      current->_leaves.push_back(tb->_leaves[i]);
      current->_leavesWeights.push_back(bb->_leavesWeights[j] + (1ULL << current->_exponent));
      i++;
      j++;
    } else if (tb->_leaves[i] <
                bb->_leaves[j]) { // Literal only in top bucket
      current->_leaves.push_back(tb->_leaves[i]);
      current->_leavesWeights.push_back(1ULL << current->_exponent);
      i++;
    } else { // Literal only in bottom bucket
      current->_leaves.push_back(bb->_leaves[j]);
      current->_leavesWeights.push_back(bb->_leavesWeights[j]);
      j++;
    }
  }
  if (_dgpwSetting->verbosity > 10)
    std::cout << "after (almost) combining the two vectors..." << std::endl;

  while (i < tb->_leaves.size()) {
    current->_leaves.push_back(tb->_leaves[i]);
    current->_leavesWeights.push_back(1ULL << current->_exponent);
    i++;
  }
  if (_dgpwSetting->verbosity > 10) {
    std::cout << "after (almost) combining the two vectors... i, j: " << i
              << ", " << j << std::endl;
    std::cout << "BB leaves size :" << bb->_leaves.size()
              << " TB leaves size : " << tb->_leaves.size()
              << " BB leavesWeights size: " << bb->_leavesWeights.size()
              << " TB leavesWeights size: " << tb->_leavesWeights.size()
              << " current exponent: " << current->_exponent
              << " current expValue: " << (1ULL << current->_exponent)
              << std::endl;
    std::cout << "BB Leaves:";

    for (int i = 0; i < bb->_leaves.size(); i++) {
      std::cout << " " << bb->_leavesWeights[i] << " lit(" << bb->_leaves[i]
                << ")";
    }
    std::cout << std::endl;
    std::cout << "TB Leaves:";
    for (int i = 0; i < tb->_leaves.size(); i++) {
      std::cout << " "
                << (1ULL << current->_exponent)
                << " lit(" << tb->_leaves[i] << ")";
    }
    std::cout << std::endl;
  }

  while (j < bb->_leaves.size()) {
    current->_leaves.push_back(bb->_leaves[j]);
    current->_leavesWeights.push_back(bb->_leavesWeights[j]);
    j++;
  }
  if (_dgpwSetting->verbosity > 10) {
    std::cout << "after combining the two vectors...2" << std::endl;

    std::cout << "Leaves:";
    for (int i = 0; i < current->_leaves.size(); i++) {
      std::cout << " " << current->_leavesWeights[i] << " lit(" << current->_leaves[i] << ")";
    }
    std::cout << std::endl;
  }
}

void DGPW::AddBookkeepingForPL(){
  substitution empty_sub = vPL->get_new_substitution();
  witnessTopBucketTare0.resize(GetP(), empty_sub);
  witnessTopBucketTare1.resize(GetP(), empty_sub);

  AddBookkeepingForPL_rec(GetOutputTree(), true, GetP() > 0 ?  _mainCascade->_structure[0]->_tares[0] : UINT32_MAX); //TODO-Tobias: Note that this makes that if we have only one bucket, the last bucket will contain a vector of ones in the _leavesWeights. This is overhead. Should we change that?
}

void DGPW::AddBookkeepingForPL_rec(TotalizerEncodeTree* tree, bool isBottomBucket, uint32_t lowTare){
  tree->_isBottomBucket = isBottomBucket;
  if (_dgpwSetting->verbosity > 10) {
    std::cout << __PRETTY_FUNCTION__ << std::endl;
    std::cout << "Low tare: " << lowTare << std::endl;
    std::cout << "Current size: " << tree->_size << std::endl;
    if (tree->_size == 1) {
      std::cout << "Leave: " << tree->_encodedOutputs[0] << std::endl;
    }
  }

  if (tree->_isBottomBucket && tree->_child1 && tree->_child1->_everyNthOutput == 2) {

    if (tree->_child1->_leaves.size() == 0)
      AddBookkeepingForPL_rec(tree->_child1, true, lowTare);
    if (tree->_child2->_leaves.size() == 0)
      AddBookkeepingForPL_rec(tree->_child2, false, lowTare);

    if (_dgpwSetting->verbosity > 10) {
      std::cout << "Visiting node " << this << std::endl;
      std::cout << "Child 1 is bottom bucket. " << std::endl;
      std::cout << "_isBottomBucket = " << tree->_isBottomBucket
                << " _child1->_isBottomBucket = " << tree->_child1->_isBottomBucket
                << "_child2->_isBottomBucket = " << tree->_child2->_isBottomBucket
                << std::endl;
    }
    

    if (tree->_exponent == UINT32_MAX)
      tree->_exponent = tree->_child1->_exponent + 1;

    CombineLeavesForBottomBucket(tree, tree->_child1, tree->_child2);

  } else if (tree->_isBottomBucket && tree->_child2 && tree->_child2->_everyNthOutput == 2) {

    if (tree->_child1->_leaves.size() == 0)
      AddBookkeepingForPL_rec(tree->_child1, false, lowTare);
    if (tree->_child2->_leaves.size() == 0)
      AddBookkeepingForPL_rec(tree->_child2, true, lowTare);

    
    if (_dgpwSetting->verbosity > 10) {
      std::cout << "Visiting node " << this << std::endl;
      std::cout << "Child 2 is bottom bucket. " << std::endl;
      std::cout << "_isBottomBucket = " << tree->_isBottomBucket
                << " _child1->_isBottomBucket = " << tree->_child1->_isBottomBucket
                << "_child2->_isBottomBucket = " << tree->_child2->_isBottomBucket
                << std::endl;
    }

    if (tree->_exponent == UINT32_MAX)
      tree->_exponent = tree->_child2->_exponent + 1;

    CombineLeavesForBottomBucket(tree, tree->_child2, tree->_child1);

  } else {

    if (tree->_child1 && tree->_child1->_leaves.size() == 0)
      AddBookkeepingForPL_rec(tree->_child1, false, lowTare);
    if (tree->_child2 && tree->_child2->_leaves.size() == 0)
      AddBookkeepingForPL_rec(tree->_child2, false, lowTare);

    
    if (_dgpwSetting->verbosity > 10) {
      std::cout << "Visiting node " << this << std::endl;
      std::cout << "We are in a top bucket" << std::endl;
    }

    if (tree->_child1) {
      if (_dgpwSetting->verbosity > 10) {
        std::cout << "child1: " << tree->_child1 << std::endl;
        std::cout << "Leaves child1: ";
      }
      for (auto leaf : tree->_child1->_leaves) {
        if (_dgpwSetting->verbosity > 10)
          std::cout << leaf << ", ";
        tree->_leaves.push_back(leaf);
      }
      if (_dgpwSetting->verbosity > 10)
        std::cout << std::endl;
    }
    if (tree->_child2) {
      if (_dgpwSetting->verbosity > 10) {
        std::cout << "child2: " << tree->_child2 << std::endl;
        std::cout << "Leaves child2: ";
      }
      for (auto leaf : tree->_child2->_leaves) {
        if (_dgpwSetting->verbosity > 10)
          std::cout << leaf << ", ";
        tree->_leaves.push_back(leaf);
      }
      if (_dgpwSetting->verbosity > 10)
        std::cout << std::endl;
    }

    if (!tree->_child1 && !tree->_child2) { // We are in a leaf!
      if (_dgpwSetting->verbosity > 10) {
        std::cout << "Indeed, we are in a leaf. " << std::endl;
        assert(tree->_size == 1);
        std::cout << "_leaves.size() = " << tree->_leaves.size() << std::endl;
      }

      if (tree->_encodedOutputs[0] < lowTare)
        tree->_leaves.push_back(tree->_encodedOutputs[0] << 1 ^ 1);
      else
        tree->_tares.push_back(tree->_encodedOutputs[0]);
      if (_dgpwSetting->verbosity > 10) {
        std::cout << "Leaves: ";
        for (auto leaf : tree->_leaves) {
          std::cout << leaf << ", ";
        }
        std::cout << std::endl;
      }
    }

    if (tree->_isBottomBucket) {
      if (_dgpwSetting->verbosity > 10)
        std::cout << "We are at the 2^0 Bottom Bucket. " << std::endl;
      // then we are in the top bottom bucket with exponent 0
      std::sort(tree->_leaves.begin(), tree->_leaves.end());
      if (tree->_exponent == UINT32_MAX)
        tree->_exponent = 0;
      for (int i = 0; i < tree->_leaves.size(); i++) {
        tree->_leavesWeights.push_back(1ULL << tree->_exponent);
      }
    }
  }

  if (tree->_encodedOutputs.size() > 1) {
    if (_dgpwSetting->verbosity > 10)
      std::cout << "Starting to copy the tares. " << std::endl;
    std::copy(tree->_child1->_tares.begin(), tree->_child1->_tares.end(),
              std::back_inserter(tree->_tares));
    std::copy(tree->_child2->_tares.begin(), tree->_child2->_tares.end(),
              std::back_inserter(tree->_tares));
  }
  if (_dgpwSetting->verbosity > 10)
    std::cout << "End visiting Node." << std::endl;
}

uint32_t DGPW::GetBucketFromTare(uint32_t tare){
  assert(GetP() > 0);
  return tare - _mainCascade->_structure[0]->_tares[0];
}

void DGPW::AddShadowLitTopBucket(VeriPB::Lit& litOrig, VeriPB::Lit& shadowLit, bool tareValue, uint32_t bucket ){
  assert(shadowLit.v.only_known_in_proof);

  #ifndef NDEBUG
    vPL->write_comment("Add to witness: " + vPL->var_name(litOrig.v) + " -> " + vPL->to_string(shadowLit) + " (tareValue = " + std::to_string(tareValue) + ")"   );
  #endif

  substitution* sub = tareValue ? &witnessTopBucketTare1[bucket] : &witnessTopBucketTare0[bucket];
  vPL->add_literal_assignment(*sub, litOrig.v, shadowLit);
}

void DGPW::SetVariableInWitnessForTopBucket(VeriPB::Var& varOrig, bool value, bool tareValue, uint32_t bucket){
  #ifndef NDEBUG
    vPL->write_comment("Add to witness: " + vPL->var_name(varOrig) + " -> " + std::to_string(value) + " (tareValue = " + std::to_string(tareValue) + ")"   );
  #endif

  substitution* sub = tareValue ? &witnessTopBucketTare1[bucket] : &witnessTopBucketTare0[bucket];
  vPL->add_boolean_assignment(*sub, varOrig, value);
}


void DGPW::CreateShadowLitsForCountingLitTopBucket(uint32_t countinglit, uint64_t reif_RHS, TotalizerEncodeTree* node){
  assert(node->_tares.size() == 1 && !node->_isBottomBucket);
  #ifndef NDEBUG
    vPL->write_comment("Top Bucket: Introducing Shadow Circuit Constraints for variable: " + vPL->to_string(countinglit) + " reif_RHS = " + std::to_string(reif_RHS));
  #endif 
  // Reification constraints
  VeriPB::Var vprime0 = vPL->new_variable_only_in_proof();
  VeriPB::Var vprime1 = vPL->new_variable_only_in_proof();
  VeriPB::Lit lprime0 = create_literal(vprime0, true);
  VeriPB::Lit lprime1 = create_literal(vprime1, true);

  std::vector<VeriPB::Lit> litsC; std::vector<wght> wghtsC; uint64_t sumOfWghts = node->_leaves.size() + 1; // + 1 because of the tare variable.
  litsC.resize(node->_leaves.size()+1); wghtsC.resize(node->_leaves.size()+1);
  
  uint32_t indexCountingLit = node->_leaves.size();
  if(reif_RHS > 1 || reif_RHS < sumOfWghts){
    for(int i = 0; i < indexCountingLit; i++){
      litsC[i] = toVeriPbLit(node->_leaves[i]);
      wghtsC[i] = 1;
    }      
    wghtsC[indexCountingLit] = reif_RHS;
  }
  
  substitution w;
  

  // Right reification
  // For t = 0    
  if(reif_RHS > 1){ 
    #ifndef NDEBUG   
    vPL->write_comment("reif right for " + vPL->to_string(countinglit) + " t = 0");
    #endif
    litsC[indexCountingLit] = neg(lprime0);

    w = vPL->get_new_substitution();
    vPL->add_boolean_assignment(w, vprime0, true);

    vPL->redundanceBasedStrengthening(litsC, wghtsC, wghtsC[indexCountingLit]-1, w);
  }

  // For t = 1
  if(reif_RHS < sumOfWghts){
    #ifndef NDEBUG
    vPL->write_comment("reif right for " + vPL->to_string(countinglit) + " t = 1");
    #endif
    litsC[indexCountingLit] = neg(lprime1);

    w = vPL->get_new_substitution();
    vPL->add_boolean_assignment(w, vprime1, true);

    vPL->redundanceBasedStrengthening(litsC, wghtsC, wghtsC[indexCountingLit], w);
  }

  if(reif_RHS > 1 || reif_RHS < sumOfWghts){
    for(int i = 0; i < litsC.size()-1; i++){
      litsC[i].negated = !litsC[i].negated;  
    }
  }

  // Left reification
  // For t = 0
  if(reif_RHS > 1){
    #ifndef NDEBUG
    vPL->write_comment("reif left for " + vPL->to_string(countinglit) + " t = 0");
    #endif

    litsC[indexCountingLit] = lprime0;
    wghtsC[indexCountingLit] = sumOfWghts - reif_RHS + 1;
    
    w = vPL->get_new_substitution();
    vPL->add_boolean_assignment(w, vprime0, false);

    vPL->redundanceBasedStrengthening(litsC, wghtsC, wghtsC[indexCountingLit], w);
  }

  // For t = 1
  if(reif_RHS < sumOfWghts){
    #ifndef NDEBUG
    vPL->write_comment("reif left for " + vPL->to_string(countinglit) + " t = 1");
    #endif

    litsC[indexCountingLit] = lprime1;
    wghtsC[indexCountingLit] = sumOfWghts - reif_RHS + 1;
    
    w = vPL->get_new_substitution();
    vPL->add_boolean_assignment(w, vprime1, false);

    vPL->redundanceBasedStrengthening(litsC, wghtsC, wghtsC[indexCountingLit]-1, w);
  }

  // Keep track of the shadow literals
  uint32_t bucket = GetBucketFromTare(node->_tares[0]);
  VeriPB::Lit countingLitV = toVeriPbLit(countinglit);
  VeriPB::Var countingVarV = variable(countingLitV);

  if(reif_RHS == 1){
    #ifndef NDEBUG
    vPL->write_comment("RHS = 1 for variable " + vPL->var_name(countingVarV));
    #endif
    SetVariableInWitnessForTopBucket(countingVarV, 0, 0, bucket);
    vPL->add_boolean_assignment(witnessT, variable(countingLitV), 0);
  }
  else{
    lprime0.negated = false;
    AddShadowLitTopBucket(countingLitV, lprime0, 0, bucket);
    vPL->add_literal_assignment(witnessT, variable(countingLitV), lprime0);
  }

  if(reif_RHS >= sumOfWghts){
    #ifndef NDEBUG
    vPL->write_comment("RHS = sumOfWghts for variable " + vPL->var_name(countingVarV) + " sumOfWghts = " + std::to_string(sumOfWghts));
    #endif
    SetVariableInWitnessForTopBucket(countingVarV, 1, 1, bucket);
  }
  else{
    lprime1.negated = false;
    AddShadowLitTopBucket(countingLitV, lprime1, 1, bucket); 
  }
}


// TODO: Creates the same vectors (CLits, CWghts) every call to this function, while it can be the same over multiple calls.
void DGPW::CreateShadowLitForBottomBucket(substitution& w, uint32_t countingVar, uint64_t reif_RHS, uint64_t T, const TotalizerEncodeTree* node){
  VeriPB::Var encodedVar = toVeriPbVar(countingVar);

  VeriPB::Var shadowvar = vPL->new_variable_only_in_proof();
  VeriPB::Lit shadowlit = create_literal(shadowvar, false);
  VeriPB::Lit shadowlitneg = create_literal(shadowvar, true);

  std::vector<VeriPB::Lit> CLits; std::vector<wght> CWghts; 
  #ifndef NDEBUG
  vPL->write_comment("[BottomBucket] Encoding shadow literal  " + vPL->to_string(shadowlitneg) + " for " + vPL->to_string(create_literal(encodedVar, true)));
  #endif
  // ~z_k <-> O' + ~T >= (k+1) * 2^exp <-> O'  >= (k+1) * 2^exp - ~T

  wght negT = (1ULL << node->_tares.size()) - 1 - T;
  
  
  wght rhsReifMinNegT = negT > reif_RHS ? 0 : reif_RHS - negT;

  // Note that the VeriPB proof checker always translates a constraint with negative left hand side to a constraint with rhs = 0;
  #ifndef NDEBUG
  vPL->write_comment("negT = " + std::to_string(negT) + " and rhs = " + std::to_string(reif_RHS));
  #endif
  // Right Implication
  CLits.push_back(shadowlit); CWghts.push_back(reif_RHS);
  for(int i = 0; i < node->_leaves.size(); i++){
    CLits.push_back(toVeriPbLit(node->_leaves[i]));
    CWghts.push_back(node->_leavesWeights[i]);
  }
  substitution witness = vPL->get_new_substitution();
  vPL->add_boolean_assignment(witness, shadowvar, true);
  constraintid reifright = vPL->redundanceBasedStrengthening(CLits, CWghts, rhsReifMinNegT, witness);
  #ifndef NDEBUG 
  vPL->write_comment("Add reif right for " + vPL->var_name(shadowvar));
  #endif
  // if(node->_isRoot) 
    vPL->setReifiedConstraintRightImpl(shadowvar, reifright);

  // Left Implication
  wght sumWghtLeaves = 0;
  for(int i = 0; i < CLits.size(); i++) {
    CLits[i] = neg(CLits[i]);
    if(i != 0)
      sumWghtLeaves += CWghts[i];
  }

  #ifndef NDEBUG
  vPL->write_comment("sumWghtLeaves " + std::to_string(sumWghtLeaves) + " (1ULL << (tree->_tares.size())) " + std::to_string((1ULL << (node->_tares.size()))) + " rhsReif " + std::to_string(reif_RHS));
  #endif
  wght rhsLeftReif = sumWghtLeaves + (1ULL << (node->_tares.size())) - reif_RHS ;
  wght rhsLefReifMinT = T > rhsLeftReif ? 0 : rhsLeftReif - T;

  CWghts[0] = rhsLeftReif;             

  witness = vPL->get_new_substitution();
  vPL->add_boolean_assignment(witness, shadowvar, false);
  constraintid reifleft = vPL->redundanceBasedStrengthening(CLits, CWghts, rhsLefReifMinT, witness);
  #ifndef NDEBUG 
  vPL->write_comment("Add reif left for " + vPL->var_name(shadowvar));
  #endif
  // if(node->_isRoot) 
    vPL->setReifiedConstraintLeftImpl(shadowvar, reifleft);
  
  // add to substitution
  vPL->add_literal_assignment(w, encodedVar, shadowlit);
}

void DGPW::SetTinWitness(substitution& w, uint64_t s, const std::vector<uint32_t>& tares){

  std::string comment = "Tares: "; 
  for(auto tare : tares) comment += vPL->var_name(tare) + " ";
  vPL->write_comment(comment);

  comment = "Witness: Set T = " + std::to_string(s) + " ";

  for(int i = GetP()-1; i>=0; i--){
    wght m = (1ULL << i);

    if(s >= m){
      s -= m;
      vPL->add_boolean_assignment(w, tares[GetP() - 1 - i], 1);
      comment += vPL->var_name(tares[i]) + " -> 1 "; 
    }
    else{
      vPL->add_boolean_assignment(w, tares[GetP() - 1 - i], 0);
      comment += vPL->var_name(tares[i]) + " -> 0 ";
    }
  }
  vPL->write_comment(comment);
}

uint32_t DGPW::GetOutputLiteral(uint32_t position){
  return _mainCascade->_structure[_mainCascade->_structure.size()-1]->_sorter->_outputTree->_encodedOutputs[position] << 1 ^ 1 ; // todo; make literal!
}

uint32_t DGPW::GetSizeOutputs(){
  #ifndef NDEBUG
  vPL->write_comment("size of outputlits: " + std::to_string(_mainCascade->_structure[_mainCascade->_structure.size()-1]->_sorter->_outputTree->_encodedOutputs.size()));
  #endif
  return _mainCascade->_structure[_mainCascade->_structure.size()-1]->_sorter->_outputTree->_encodedOutputs.size();
}

uint32_t DGPW::GetMaxPos() {
  return _mainCascade->_maxPos;
}

void DGPW::GetTares(std::vector<uint32_t>& tares) {
  tares.clear();
  for ( auto bucket : _mainCascade->_structure ) {
    if ( bucket->_tares.empty() ) continue;
    tares.push_back(bucket->_tares[0]);
  }
}

void DGPW::GetTares(std::vector<VeriPB::Lit>& tares, bool taresNegated){
  tares.clear();
  for ( auto bucket : _mainCascade->_structure ) {
    if ( bucket->_tares.empty() ) continue;
    uint32_t tare = create_literal(bucket->_tares[0], taresNegated);
    tares.push_back(toVeriPbLit(tare));
  }
}

// PROOF LOGGING: Creation of shadow circuit in the proof
void DGPW::CreateShadowCircuitPL(uint64_t s, substitution& w, constraintid cxnLBcurrentGBMO){
  assert(s > 0); // Shadow circuit for T = 0 is derived while creating the encoding already. 
    
  #ifndef NDEBUG
  vPL->write_comment("Creation of shadow circuit for T = " + std::to_string(s));
  #endif

  TotalizerEncodeTree* tree = GetOutputTree();

  SetTinWitness(w, s, tree->_tares);

  std::unordered_map<uint32_t, uint64_t> valuesTareVariables; // tare[var] = 2^i if variable needs to be assigned 1 such that T = s or 0 otherwise.
  valuesTareVariables.reserve(GetP()); 


  // Add the witnesses for the top bucket variables.
  for(int i = GetP()-1; i>=0; i--){
    wght m = (1ULL << i);

    uint32_t tare = _mainCascade->_structure[i]->_tares[0];
    uint32_t bucket = GetBucketFromTare(tare);

    if(s >= m){
      valuesTareVariables[tare] = m;
      vPL->add_substitution(w, witnessTopBucketTare1[bucket]);
      s -= m;
    }
    else{
      valuesTareVariables[tare]  = 0;
      vPL->add_substitution(w, witnessTopBucketTare0[bucket]);   
    }
  }

  // Print comment for debugging reasons:
  #ifndef NDEBUG
  std::string contentofvaluesTareVariables = "Tares: "; 
  for(const auto& pair : valuesTareVariables){
    if(pair.second > 0)
      contentofvaluesTareVariables += vPL->var_name(pair.first) + " -> 1 ";
    else 
      contentofvaluesTareVariables += vPL->var_name(pair.first) + " -> 0 ";
  }
  vPL->write_comment(contentofvaluesTareVariables);
  contentofvaluesTareVariables = "Values for Tares: ";
  for(const auto& pair : valuesTareVariables){
    contentofvaluesTareVariables += vPL->var_name(pair.first) + " (" + std::to_string(pair.first) + ") -> " + std::to_string(pair.second) + " ";
  }
  vPL->write_comment(contentofvaluesTareVariables);
  vPL->write_comment("Shadow Circuit creation - Start traversing the tree for T = " + std::to_string(s));
  #endif
  // Create the shadow circuit for the bottom buckets recursively by traversing the encoding tree
  CreateShadowCircuitPL_rec(w, tree,valuesTareVariables, true); 
  vPL->write_comment("Shadow Circuit creation - End traversing the tree");
}

void DGPW::CreateSubproofsAlreadySatisfiedShadowedLits(std::vector<subproof>& subproofs, cuttingplanes_derivation cpderCxnLBcurrentGBMO, substitution& w, uint32_t varforproofgoalhash1){
  cuttingplanes_derivation cpder;
  cuttingplanes_derivation cpdertoapply = CreateCPDerForSubproofAlreadySatisfiedLit(_pacose->_GCD, cpderCxnLBcurrentGBMO);
  vPL->write_comment("Creating subproofs for already satisfied shadowed lits");
  for(auto p : CoarseConvergenceCxnidForSAT){
    uint32_t originalVar = p.first;
    std::string proofgoal = std::to_string(p.second);

    #ifndef NDEBUG
    vPL->write_comment("Proof goal " + proofgoal + " for literal " + vPL->var_name(originalVar));
    #endif
    subproofs.push_back(CreateSubproofAlreadySatisfiedShadowedLit(proofgoal, w, originalVar, cpdertoapply));
  }

  if(varforproofgoalhash1 != 0){
    #ifndef NDEBUG
    vPL->write_comment("Proof goal #1 for literal " + vPL->var_name(varforproofgoalhash1));
    #endif
    subproofs.push_back(CreateSubproofAlreadySatisfiedShadowedLit("#1", w, varforproofgoalhash1, cpdertoapply));
  }
}

subproof DGPW::CreateSubproofAlreadySatisfiedShadowedLit(std::string proofgoal, substitution& w, uint32_t& shadowedvar, cuttingplanes_derivation& cpdertoapply){
  #ifndef NDEBUG
  vPL->write_comment("Proof goal " + proofgoal + " for literal " + vPL->var_name(shadowedvar));
  #endif
  cuttingplanes_derivation cpder = vPL->CP_constraintid(vPL->getReifiedConstraintLeftImpl(variable(vPL->get_literal_assignment(w, shadowedvar))));
  cpder = vPL->CP_apply(cpder, cpdertoapply);

  std::vector<cuttingplanes_derivation> cpders; cpders.push_back(cpder);
  return {proofgoal, cpders};
}

cuttingplanes_derivation DGPW::CreateCPDerForSubproofAlreadySatisfiedLit(uint64_t GCD, cuttingplanes_derivation& cpderCxnLBcurrentGBMO){
  return vPL->CP_addition(vPL->CP_saturation( vPL->CP_division(vPL->CP_addition(vPL->CP_multiplication(GCD), cpderCxnLBcurrentGBMO), GCD) ), vPL->CP_constraintid(-1));
}

void DGPW::SetProofgoalLastSubproof(std::vector<subproof>& subproofs, std::string proofgoal){
  subproofs[subproofs.size()-1].proofgoal = proofgoal;
}

void DGPW::CreateShadowCircuitPL_rec(substitution& w, const TotalizerEncodeTree* tree, const std::unordered_map<uint32_t, uint64_t>& valuesTareVariables, bool is_root){
  if(!tree->_isBottomBucket || tree->_tares.size() == 0) // We only need to create the shadow circuit for bottom buckets.
        return;
  
  #ifndef NDEBUG
  uintptr_t node_id = reinterpret_cast<uintptr_t>(tree);
  std::string leavesstr = std::to_string(node_id) + " - Leaves: ";
  for(int i = 0; i < tree->_leaves.size(); i++){
    if(tree->_leavesWeights.size() > 0)
      leavesstr += "(" + vPL->to_string(tree->_leaves[i]) + "," + std::to_string(tree->_leavesWeights[i]) +  ") ";
    else
      leavesstr += vPL->to_string(tree->_leaves[i]) + " ";
    
  }
  std::string encodedoutputsstr = "EncodedOutputs: ";
  for(auto output : tree->_encodedOutputs){
    encodedoutputsstr += vPL->var_name(output) + " ";
  }
  std::string taresstr = "Tares: ";
  for(auto tare : tree->_tares){
    taresstr += vPL->var_name(tare) + " ";
  }
  vPL->write_comment(leavesstr + encodedoutputsstr + taresstr);
  #endif

  wght T = 0;
  for(auto tare : tree->_tares){
    assert(valuesTareVariables.find(tree->_tares[0]) != valuesTareVariables.end());
    T += valuesTareVariables.at(tare);
  }

  wght negT = (1ULL << tree->_tares.size()) - 1 - T;

  // #ifndef NDEBUG
  //   std::cout << "tree->_encodedOutputs: ";
  //   for(auto var : tree->_encodedOutputs) std::cout << " " << vPL->var_name(var);
  //   std::cout << std::endl;
  // #endif

  for(uint32_t k = 0; k < tree->_encodedOutputs.size(); k++){
    // Call CreateShadowLitForBottomBucket for reif_RHS = (k + 1) * 2^exp

    if(tree->_encodedOutputs.size() > 0 && tree->_encodedOutputs[k] == 0){ 
      vPL->write_comment("Encoded outputs == 0");
      continue; // Current variable was not derived yet
    }

    wght rhsReif = (k+1)*(1ULL << tree->_exponent);

    CreateShadowLitForBottomBucket(w, tree->_encodedOutputs[k], rhsReif, T, tree);
  }

  vPL->write_comment("Done creating shadow circuit for node.");

  if(tree->_child1 != nullptr) CreateShadowCircuitPL_rec(w, tree->_child1, valuesTareVariables, false);
  if(tree->_child2 != nullptr) CreateShadowCircuitPL_rec(w, tree->_child2, valuesTareVariables, false);
}

// Functions for Proof Logging
constraintid DGPW::derivelbT(uint64_t lb, TotalizerEncodeTree* tree, substitution witnessT, std::vector<subproof>& subproofs){
  #ifndef NDEBUG
  vPL->write_comment("T >= " + std::to_string(lb));
  #endif
  std::vector<uint32_t> Clits; std::vector<uint64_t> Cwghts;
  for(int i = 0; i < tree->_tares.size(); i++){
      Clits.push_back(create_literal(tree->_tares[i], false));
      vPL->write_comment("Variable " + vPL->var_name(tree->_tares[i]) + " with weight " + std::to_string(1ULL << (tree->_tares.size() - 1 -  i )));
      Cwghts.push_back(1ULL << (tree->_tares.size() - 1 -  i ));
  }
  cxnlbT = vPL->redundanceBasedStrengthening(Clits, Cwghts, lb, witnessT, subproofs);
  return cxnlbT;
}

constraintid DGPW::deriveubT(uint64_t ub, TotalizerEncodeTree* tree, substitution witnessT, std::vector<subproof>& subproofs){
  #ifndef NDEBUG
  vPL->write_comment("Derive T =< " + std::to_string(ub));
  #endif
  std::vector<uint32_t> Clits; std::vector<uint64_t> Cwghts;
  for(int i = 0; i < tree->_tares.size(); i++){
      Clits.push_back(create_literal(tree->_tares[i], true));
      Cwghts.push_back(1ULL << (tree->_tares.size() - 1 -  i ));
  }
  uint64_t p = GetP();
  cxnubT = vPL->redundanceBasedStrengthening(Clits, Cwghts, (1ULL << p) - 1 -  ub , witnessT, subproofs);
  // cxnubT = vPL->rup(Clits, Cwghts, (1ULL << p) - 1 -  ub);
  return cxnubT;
}

}  // namespace DGPW
} // Namespace Pacose
